#include "CellML_Context.hh"
#include "ICellML_Context.h"
#include <nsIFile.h>
#include <stdio.h>
#include <nsStringAPI.h>
#include <nsGenericFactory.h>
#include <nsIComponentRegistrar.h>
#include <nsILineInputStream.h>
#include <nsIInputStream.h>
#include <Icorba4xpcom.h>
#include <nsIProcess.h>
#include <nsNetCID.h>
#include <nsIFileStreams.h>
#include <nsComponentManagerUtils.h>
#include <nsServiceManagerUtils.h>
#include <nsIProperties.h>

#define CELLML_CONTEXT_CID                           \
{ /* 70640b44-c537-4b4e-a584-43c0bdef74ee */         \
    0x70640b44,                                      \
    0xc537,                                          \
    0x4b4e,                                          \
    {0xa5, 0x84, 0x43, 0xc0, 0xbd, 0xef, 0x74, 0xee} \
}

#include "ThreadUtilities.h"

static NS_IMETHODIMP
AttemptContactCellMLContext
(
 nsIFile* aIORFile,
 cellml_context_ICellMLContext** aCtxt
)
{
  nsCOMPtr<nsIInputStream> strm;
  nsresult rv;
  nsCOMPtr<nsILineInputStream> lstrm =
    do_CreateInstance(NS_LOCALFILEINPUTSTREAM_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  if (lstrm == nsnull)
  {
    return NS_ERROR_FAILURE;
  }
  PRBool more;
  nsCString buffer;

  nsCOMPtr<nsIFileInputStream> fstrm =
    do_QueryInterface(lstrm);

  rv = fstrm->Init(aIORFile, -1, -1, 0);
  NS_ENSURE_SUCCESS(rv, rv);

  rv = lstrm->ReadLine(buffer, &more);
  NS_ENSURE_SUCCESS(rv, rv);

  // Fetch the ORB proxy...
  nsCOMPtr<CORBA_IORBProxy> iop(do_GetService("@cellml.org/CORBA/ORB;1"));
  if (iop == nsnull)
  {
    return NS_ERROR_FAILURE;
  }

  // Make sure the CellML XPCORBA module is in...
  PRBool retval;
  iop->Load_module("{ec337bea-b97f-4be2-8bc9-2b676146e43c}", "cellml",
                   &retval);

  // If we get here, we have an IOR, but it could well be stale...
  nsCOMPtr<nsISupports> obj;

  rv = iop->String_to_object(buffer.BeginReading(), getter_AddRefs(obj));
  // If it is invalid, we return an error, so we can retry...
  NS_ENSURE_SUCCESS(rv, rv);

  // We have an object, but it might not be still responding. Performing a
  // QI will tell us, however...
  rv = obj->QueryInterface(NS_GET_IID(cellml_context_ICellMLContext),
                           (void**)aCtxt);
  NS_ENSURE_SUCCESS(rv, rv);

  return NS_OK;
}

static NS_IMETHODIMP
CellMLContextConstructor(nsISupports* aOuter, REFNSIID aIID, void** aResult)
{
  if (aOuter != nsnull)
  {
    printf("CellML Context: Cannot aggregate with another object.\n");
    return NS_ERROR_NO_AGGREGATION;
  }

  // Find the user's home directory...
  nsCOMPtr<nsIProperties> props
    (do_GetService("@mozilla.org/file/directory_service;1"));

  if (props == nsnull)
  {
    printf("CellML Context: Can't get the directory service!\n");
    return NS_ERROR_OUT_OF_MEMORY;
  }

  nsresult ret;
  nsCOMPtr<nsIFile> homedir;
#ifdef  XP_WIN
  ret = props->Get("LocalAppData", NS_GET_IID(nsIFile), getter_AddRefs(homedir));
#else
  ret = props->Get("Home", NS_GET_IID(nsIFile), getter_AddRefs(homedir));
#endif
  NS_ENSURE_SUCCESS(ret, ret);
  if (homedir == nsnull)
  {
    printf("CellML Context construction: Home directory fetch failed.\n");
    return NS_ERROR_OUT_OF_MEMORY;
  }

  nsCOMPtr<nsIFile> iorFile;
  ret = homedir->Clone(getter_AddRefs(iorFile));
  NS_ENSURE_SUCCESS(ret, ret);
  if (iorFile == nsnull)
  {
    printf("CellML Context construction: Home directory clone failed.\n");
    return NS_ERROR_OUT_OF_MEMORY;
  }

  ret = iorFile->Append(NS_LITERAL_STRING(".cellml"));
  NS_ENSURE_SUCCESS(ret, ret);
  ret = iorFile->Append(NS_LITERAL_STRING("corba_server"));
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMPtr<cellml_context_ICellMLContext> ctxt;
  ret = AttemptContactCellMLContext(iorFile, getter_AddRefs(ctxt));
  
  if (ret == NS_OK)
  {
    return ctxt->QueryInterface(aIID, aResult);
  }

  // There is no live server. Create a process...
  nsCOMPtr<nsIProcess> serverProcess =
    do_CreateInstance(NS_PROCESS_CONTRACTID, &ret);
  NS_ENSURE_SUCCESS(ret, ret);

  // Find our version of the context server...
  nsCOMPtr<nsIFile> profdir;
  ret = props->Get("ProfD", NS_GET_IID(nsIFile), getter_AddRefs(profdir));
  NS_ENSURE_SUCCESS(ret, ret);
  if (profdir == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;
  ret = profdir->Append(NS_LITERAL_STRING("extensions"));
  NS_ENSURE_SUCCESS(ret, ret);
  ret = profdir->Append(NS_LITERAL_STRING("{ec337bea-b97f-4be2-8bc9-2b676146e43c}"));
  NS_ENSURE_SUCCESS(ret, ret);
  ret = profdir->Append(NS_LITERAL_STRING("components"));
  NS_ENSURE_SUCCESS(ret, ret);
  ret = profdir->Append(NS_LITERAL_STRING("cellml_corba_server"));
  NS_ENSURE_SUCCESS(ret, ret);
  PRBool exist;
  ret = profdir->Exists(&exist);
  NS_ENSURE_SUCCESS(ret, ret);

  // No context server installed here, and none running already.
  if (!exist)
    return NS_ERROR_FAILURE;

  // Initialise the server...
  ret = serverProcess->Init(profdir);
  NS_ENSURE_SUCCESS(ret, ret);
  PRUint32 pid;
  ret = serverProcess->Run(PR_FALSE, nsnull, 0, &pid);
  NS_ENSURE_SUCCESS(ret, ret);

  // The IOR will appear sooner or later, and we just have to wait for it. This
  // is a little bit ugly, because it is blocking, but since this is the
  // bootstrap we have no other means to know the server is ready...
  PRUint32 i;
  for (i = 0; i < 10; i++)
  {
    PR_Sleep(PR_MillisecondsToInterval(200));

    ret = AttemptContactCellMLContext(iorFile, getter_AddRefs(ctxt));
    
    if (ret == NS_OK)
    {
      return ctxt->QueryInterface(aIID, aResult);
    }
  }

  return NS_ERROR_FAILURE;
}

NS_GENERIC_FACTORY_CONSTRUCTOR(ThreadUtilities)

static const nsModuleComponentInfo components[] = 
{
  {
    "CellMLContext",
    CELLML_CONTEXT_CID,
    "@cellml.org/cellml-context;1",
    CellMLContextConstructor
  },
  {
    "ThreadUtilities",
    THREAD_UTILITIES_CID,
    "@cellml.org/thread-utilities;1",
    ThreadUtilitiesConstructor
  }
};

NS_IMPL_NSGETMODULE(CellMLBootstrap,components)
